#include <sys/types.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#define LOGBUFSIZE 32678

static FILE *fp;
static const char *logfile_basename;
static int logfile_generations=10;
static int logfile_maxsize=1000000;

static void parse_arguments(int argc, char **argv) {
	int c;
	while((c=getopt(argc,argv,"vg:s:"))!=-1) {
		switch(c) {
			case 'v':
				printf("hplog - High Performance Logger v1.0\n");
				exit(1);
			case 'g':
				logfile_generations = (int)strtol(optarg,0,0);
				if(logfile_generations<0)
					logfile_generations = 0;
				break;
			case 's':
				logfile_maxsize = (int)strtol(optarg,0,0);
				if(logfile_maxsize<0)
					logfile_maxsize = 0;
				break;
		}
	}
	if(optind<argc)
		logfile_basename = argv[optind];
	else
		logfile_basename = "log";
}

static void openlogfile() {
	fp = fopen(logfile_basename,"a");
	if(!fp) {
		perror("hplog: could not open log file");
		exit(-1);
	}
	setvbuf(fp,0,_IOFBF,LOGBUFSIZE);
}
static void closelogfile() {
	fflush(fp);
	fclose(fp);
}
static void rotatelogfiles() {
	char tmp1[512];
	char tmp2[512];
	int i;
	sprintf(tmp1,"%s.%d",logfile_basename,logfile_generations);
	remove(tmp1);
	for(i=logfile_generations; i>0; i--) {
		sprintf(tmp1,"%s.%d",logfile_basename,i-1);
		sprintf(tmp2,"%s.%d",logfile_basename,i);
		rename(tmp1,tmp2);
	}
	rename(logfile_basename,tmp1);
}

int main(int argc, char **argv) {
	time_t last_sizecheck_time = 0;
	parse_arguments(argc,argv);
	openlogfile();
	fcntl(0, F_SETFL, O_NONBLOCK);
	signal(SIGPIPE,SIG_IGN);
	
	for(;;) {
		char buf[16384];
		int bytesread;
		int should_rotate;
		int should_closeopen;
		
		bytesread = read(0,buf,sizeof(buf));
		if(bytesread<=0) {
			fd_set fds;

			if(errno!=EAGAIN && errno!=EINTR) {
				close(0);
				closelogfile();
				return 0;
			}
			fflush(fp);
			
			FD_ZERO(&fds);
			FD_SET(0,&fds);
			select(0+1, &fds,0,0, NULL);
			
			bytesread = read(0,buf,sizeof(buf));
			if(bytesread<=0 && (errno==EAGAIN || errno==EINTR)) {
				/*Seen on HP/UX - don't panic */
				continue;
			}
			if(bytesread<=0) {
				close(0);
				closelogfile();
				return 0;
			}
		}
		should_rotate=0;
		should_closeopen=0;
		if(last_sizecheck_time+1<=time(0)) {
			/*Time for checking if we should rotate log files*/
			last_sizecheck_time = time(0);
			fflush(fp);
			if(ftell(fp) >= logfile_maxsize) {
				should_rotate = 1;
				should_closeopen = 1;
			} else {
				struct stat disk_s,open_s;
				if(stat(argv[1],&disk_s)!=0 ||
				   fstat(fileno(fp),&open_s)!=0 ||
				   disk_s.st_dev!=open_s.st_dev ||
				   disk_s.st_ino!=open_s.st_ino)
				{
					/*Someone has renamed the file*/
				   	should_closeopen = 1;
				}
			}
		}
		if(should_rotate || should_closeopen) {
			char *p=(char*)memchr(buf,'\n',bytesread);
			if(!p) p=buf;
			else p++;
			
			if(p!=buf)
				fwrite(buf,p-buf,1,fp);
			
			closelogfile();

			if(should_rotate)
				rotatelogfiles();

			openlogfile();
			if(p!=buf+bytesread)
				fwrite(p,bytesread-(p-buf),1,fp);
		} else
			fwrite(buf,bytesread,1,fp);
	}
}
